'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.getScopedDispatchMethods = getScopedDispatchMethods;

var _invariant = require('invariant');

var _invariant2 = _interopRequireDefault(_invariant);

var _rx = require('rx');

var _rx2 = _interopRequireDefault(_rx);

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _redux = require('redux');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

/**
 * Wraps action creators in redux bind
 *
 * @param actionCreators
 * @returns {Function}
 */

function getScopedDispatchMethods(actionCreators) {
    return function (dispatch) {
        return (0, _redux.bindActionCreators)(actionCreators, dispatch);
    };
}

// default target state and dispatch mappers
var defaultStateMapper = function defaultStateMapper(state) {
    return state;
};
var defaultDispatchMapper = function defaultDispatchMapper(dispatch) {
    return { dispatch: dispatch };
};

/**
 * Assigns scoped state slice to the bound component.
 *
 * If bindToProperty parameter is passed,
 * scoped state will be nested on the target component.
 *
 * @param target
 * @param stateSlice
 * @param dispatch
 * @param bindToProperty
 */

function updateTarget(target, stateSlice, dispatch, bindToProperty) {

    if (bindToProperty) {
        stateSlice = _defineProperty({}, bindToProperty, stateSlice);
    }

    if (_lodash2.default.isFunction(target)) {
        target(stateSlice, dispatch);
    } else {
        _lodash2.default.merge(target, stateSlice, dispatch);
    }
}

/**
 * Returns the part of the state store that the connected component requires.
 *
 * @param state
 * @param stateMapper
 * @returns {*}
 */

function getScopedState(state, stateMapper) {

    return stateMapper(state);
}

/**
 * Connector class that connects an angular 2 app with redux.
 * Provides bindings for the flat redux api.
 *r
 * getState()
 * dispatch(action)
 * subscribe(listener)
 *
 */

function ReduxBinder(store) {

    var ReduxBinder = {
        store: store
    };

    /**
     * Instead of binding the applications properties directly, createObserver
     * supplies the component with an RXJS Observer instead that will sequelize the new state
     * on every update.
     *
     * createObservable expects two functions.
     *
     * @param stateMapper
     *
     * A function that specifies the mapping of the whole state tree to the
     * components properties.
     *
     * @param dispatchMapper
     *
     * Specifies which action creators should be directly bound to the component.
     * By default simply attaches the store's dispatch method to the component.
     *
     * To undestand this better:
     * @see http://rackt.org/redux/docs/api/bindActionCreators.html
     *
     * @returns {Function}
     */

    ReduxBinder.createObservable = function () {
        var _this = this;

        var stateMapper = arguments.length <= 0 || arguments[0] === undefined ? defaultStateMapper : arguments[0];
        var dispatchMapper = arguments.length <= 1 || arguments[1] === undefined ? defaultDispatchMapper : arguments[1];


        this.validateInputs(stateMapper, dispatchMapper);

        /**
         * Creates an observable that as a next sequence in time gets the applications
         * new state ( actually, the scopedState ofthe applications state that the bound component
         * specifies through the getScopedState method.)
         * @param store
         *
         * Redux store
         *
         * Connect returns the function to unsubscribe from the redux store.
         *
         * Redux's api for subscribe returns the unsubscribe function to the store.
         * We forward it to the component that binds itself to the store.
         *
         * We wrap this into an RX observable and return the observable instead of the unsubscribe link
         * that usually comes with the event. This way having a memory leak is not possible and
         * we dont have to worry about unsubscribing plus we get all the awesome functionality of RXJS.
         *
         * @type {Object|undefined|void|*}
         */

        return _rx2.default.Observable.create(function (observer) {

            // Make observer update with updated & scoped state slice
            // See what happens here: on every subscribe event, the observer's onNext is called.
            // TODO: think a bit more to which degree this could potentially call memory leaks
            _this.store.subscribe(function () {
                observer.onNext(getScopedState(_this.store.getState(), stateMapper));
            });
        });
    };

    /**
     * Validates mapper inputs.
     *
     * @param stateMapper
     * @param dispatchMapper
     */

    ReduxBinder.validateInputs = function (stateMapper, dispatchMapper) {

        (0, _invariant2.default)(_lodash2.default.isFunction(stateMapper), 'stateMapper parameter expects a function.');

        (0, _invariant2.default)(_lodash2.default.isPlainObject(dispatchMapper) || _lodash2.default.isFunction(dispatchMapper), 'dispatchMapper parameter expects a function or an object.');
    };

    /**
     * Binds store updates to the component so that the component
     * will always automatically update its own properties according
     * to a passed mapping.
     *
     *
     * Connect expects two functions.
     *
     * @param stateMapper
     *
     * A function that specifies the mapping of the whole state tree to the
     * components properties.
     *
     * @param dispatchMapper
     *
     * Specifies which action creators should be directly bound to the component.
     * By default simply attaches the store's dispatch method to the component.
     *
     * To undestand this better:
     * @see http://rackt.org/redux/docs/api/bindActionCreators.html
     *
     * @param bindToProperty
     * @returns {Function}
     */

    ReduxBinder.bindStore = function () {
        var stateMapper = arguments.length <= 0 || arguments[0] === undefined ? defaultStateMapper : arguments[0];

        var _this2 = this;

        var dispatchMapper = arguments.length <= 1 || arguments[1] === undefined ? defaultDispatchMapper : arguments[1];
        var bindToProperty = arguments.length <= 2 || arguments[2] === undefined ? 'appState' : arguments[2];


        this.validateInputs(stateMapper, dispatchMapper);

        var scopedState = getScopedState(this.store.getState(), stateMapper);

        if (!_lodash2.default.isFunction(dispatchMapper)) {
            dispatchMapper = getScopedDispatchMethods(dispatchMapper);
        }
        var boundActionCreators = dispatchMapper(this.store.dispatch);

        /**
         * The return value of connect is actually an unsubscribe function that needs to be called with
         * the component that should be bound to redux.
         *
         * Use like:
         *
         * class Test{
         *      constructor(){
         *
         *          // call the returned function using this
         *          this.unsubscribe = connect(func1,func2)(this)
         *      }
         *
         *      onDestroy(){
         *          this.unsubscribe();
         *      }
         * }
         *
         */

        return function (target) {

            (0, _invariant2.default)(_lodash2.default.isFunction(target) || _lodash2.default.isObject(target), 'The target parameter passed to connect must be a Function or a plain object.');

            target.appState = {};

            // Updates the target's state scopedState on initialisation with
            // initial values and dispatch actions.
            updateTarget(target, scopedState, boundActionCreators, bindToProperty);

            // Binds the target component update to the store subscription.
            return _this2.store.subscribe(function () {
                scopedState = getScopedState(_this2.store.getState(), stateMapper);
                updateTarget(target, scopedState, boundActionCreators, bindToProperty);
            });
        };
    };

    return ReduxBinder;
}

exports.default = ReduxBinder;