'use strict';

import invariant from 'invariant';
import Rx from 'rx'
import _ from 'lodash';
import { bindActionCreators } from 'redux';

/**
 * Wraps action creators in redux bind
 *
 * @param actionCreators
 * @returns {Function}
 */

export function getScopedDispatchMethods(actionCreators) {
    return dispatch => bindActionCreators(actionCreators, dispatch);
}

// default target state and dispatch mappers
const defaultStateMapper = (state) => (state);
const defaultDispatchMapper = dispatch => ({dispatch});


/**
 * Assigns scoped state slice to the bound component.
 *
 * If bindToProperty parameter is passed,
 * scoped state will be nested on the target component.
 *
 * @param target
 * @param stateSlice
 * @param dispatch
 * @param bindToProperty
 */

function updateTarget(target, stateSlice, dispatch, bindToProperty) {

    if (bindToProperty) {
        stateSlice = {
            [bindToProperty]: stateSlice
        }
    }

    if (_.isFunction(target)) {
        target(stateSlice, dispatch);
    } else {
        _.merge(target, stateSlice, dispatch);
    }
}

/**
 * Returns the part of the state store that the connected component requires.
 *
 * @param state
 * @param stateMapper
 * @returns {*}
 */

function getScopedState(state, stateMapper) {

    return stateMapper(state);
}


/**
 * Connector class that connects an angular 2 app with redux.
 * Provides bindings for the flat redux api.
 *r
 * getState()
 * dispatch(action)
 * subscribe(listener)
 *
 */

function ReduxBinder(store) {

    const ReduxBinder = {
        store
    };

    /**
     * Instead of binding the applications properties directly, createObserver
     * supplies the component with an RXJS Observer instead that will sequelize the new state
     * on every update.
     *
     * createObservable expects two functions.
     *
     * @param stateMapper
     *
     * A function that specifies the mapping of the whole state tree to the
     * components properties.
     *
     * @param dispatchMapper
     *
     * Specifies which action creators should be directly bound to the component.
     * By default simply attaches the store's dispatch method to the component.
     *
     * To undestand this better:
     * @see http://rackt.org/redux/docs/api/bindActionCreators.html
     *
     * @returns {Function}
     */

    ReduxBinder.createObservable = function (stateMapper = defaultStateMapper,
                     dispatchMapper = defaultDispatchMapper) {

        this.validateInputs(stateMapper, dispatchMapper);


        /**
         * Creates an observable that as a next sequence in time gets the applications
         * new state ( actually, the scopedState ofthe applications state that the bound component
         * specifies through the getScopedState method.)
         * @param store
         *
         * Redux store
         *
         * Connect returns the function to unsubscribe from the redux store.
         *
         * Redux's api for subscribe returns the unsubscribe function to the store.
         * We forward it to the component that binds itself to the store.
         *
         * We wrap this into an RX observable and return the observable instead of the unsubscribe link
         * that usually comes with the event. This way having a memory leak is not possible and
         * we dont have to worry about unsubscribing plus we get all the awesome functionality of RXJS.
         *
         * @type {Object|undefined|void|*}
         */

        return Rx.Observable.create(observer => {

            // Make observer update with updated & scoped state slice
            // See what happens here: on every subscribe event, the observer's onNext is called.
            // TODO: think a bit more to which degree this could potentially call memory leaks
            this.store.subscribe(() => {
                    observer.onNext(
                        getScopedState(this.store.getState(), stateMapper)
                    )
                }
            );
        });

    };

    /**
     * Validates mapper inputs.
     *
     * @param stateMapper
     * @param dispatchMapper
     */

    ReduxBinder.validateInputs = function (stateMapper, dispatchMapper) {

        invariant(
            _.isFunction(stateMapper),
            'stateMapper parameter expects a function.'
        );

        invariant(
            _.isPlainObject(dispatchMapper) || _.isFunction(dispatchMapper),
            'dispatchMapper parameter expects a function or an object.'
        );
    }

    /**
     * Binds store updates to the component so that the component
     * will always automatically update its own properties according
     * to a passed mapping.
     *
     *
     * Connect expects two functions.
     *
     * @param stateMapper
     *
     * A function that specifies the mapping of the whole state tree to the
     * components properties.
     *
     * @param dispatchMapper
     *
     * Specifies which action creators should be directly bound to the component.
     * By default simply attaches the store's dispatch method to the component.
     *
     * To undestand this better:
     * @see http://rackt.org/redux/docs/api/bindActionCreators.html
     *
     * @param bindToProperty
     * @returns {Function}
     */

    ReduxBinder.bindStore = function (stateMapper = defaultStateMapper,
              dispatchMapper = defaultDispatchMapper,
              bindToProperty = 'appState') {

        this.validateInputs(stateMapper, dispatchMapper);

        let scopedState = getScopedState(this.store.getState(), stateMapper);

        if (!_.isFunction(dispatchMapper)) {
            dispatchMapper = getScopedDispatchMethods(dispatchMapper);
        }
        const boundActionCreators = dispatchMapper(this.store.dispatch);

        /**
         * The return value of connect is actually an unsubscribe function that needs to be called with
         * the component that should be bound to redux.
         *
         * Use like:
         *
         * class Test{
         *      constructor(){
         *
         *          // call the returned function using this
         *          this.unsubscribe = connect(func1,func2)(this)
         *      }
         *
         *      onDestroy(){
         *          this.unsubscribe();
         *      }
         * }
         *
         */

        return (target) => {

            invariant(
                _.isFunction(target) || _.isObject(target),
                'The target parameter passed to connect must be a Function or a plain object.'
            );

            target.appState = {};

            // Updates the target's state scopedState on initialisation with
            // initial values and dispatch actions.
            updateTarget(target, scopedState, boundActionCreators, bindToProperty);

            // Binds the target component update to the store subscription.
            return this.store.subscribe(() => {
                scopedState = getScopedState(this.store.getState(), stateMapper);
                updateTarget(target, scopedState, boundActionCreators, bindToProperty);
            });
        }
    };

    return ReduxBinder;
}

export default ReduxBinder;