'use strict';

import chai from 'chai';
import { createStore } from 'redux';
import ReduxBinder from '../src/reduxBind/binder';
import _ from 'lodash';
import util from 'util';

const expect = chai.expect;
const assert = chai.assert;

class TestApp {
    constructor() {
        this.name = 'hello';
        this.state = {};
    }
}

describe('ReduxBinder', () => {
    let store;
    let storeBinder;
    let targetObj;
    let defaultState;
    let app;

    const dispatchMapper = null;
    const stateMapper = null;

    console.log('redux binder', ReduxBinder);

    describe('#bind', function () {


        beforeEach(() => {
            defaultState = {
                test1: 'sometest',
                test2: -1
            };

            // super simple reducer for test
            store = createStore(
                (state = defaultState, action = {}) => {
                    console.log('NEW STATE', _.merge(state, {test2: action.body}));
                    return _.merge(state, {test2: action.body});
                });
            targetObj = {};
            app = new TestApp();
            storeBinder = new ReduxBinder(store);
        });

        it('Should return an unsubscribe function', () => {
            let unsubscribe = storeBinder.bindStore()(app);
            assert.ok(unsubscribe);
            assert.isFunction(unsubscribe);
            unsubscribe();
        });


        it('Should extend target (Object) with selected state once directly after creation', () => {
            let unsubscribe = storeBinder.bindStore(
                () => ({
                    vm: {test: 1}
                }))(app);
            expect(app.appState.vm).to.deep.equal({test: 1});
            unsubscribe();
        });

        it('Should update the target (Object) passed to connect when the store updates', () => {
            let unsubscribe = storeBinder.bindStore(state => state)(app);
            store.dispatch({type: 'ACTION', body: 0});
            expect(app.appState.test2).to.equal(0);
            store.dispatch({type: 'ACTION', body: 7});
            expect(app.appState.test2).to.equal(7);
            unsubscribe();
        });

        it('Should extend target (object) with actionCreators', () => {
            storeBinder.bindStore(()=> ({}),
                () => {
                    return {
                        ac1: () => {},
                        ac2: () => {}
                    }
                }
            )(app);
            assert.isFunction(app.ac1);
            assert.isFunction(app.ac2);
        });
    });

    describe('#createObserver', function () {
        beforeEach(() => {
            defaultState = {
                test1: 'sometest',
                test2: -1
            };

            // super simple reducer for test
            store = createStore(
                (state = defaultState, action = {}) => {
                    return _.merge(state, {test2: action.body});
                });
            targetObj = {};
            app = new TestApp();
            storeBinder = new ReduxBinder(store);
        });

        it('Should return an RX observable subscription', () => {
            let source = storeBinder.createObservable();
            assert.ok(source);
        });


        it('Should enable sequentially receiving state updates', cb => {
            let source = storeBinder.createObservable(state => state);
            let counter = 0;

            source.subscribe(
                stateSlice => {
                    switch (counter) {
                        case 0:
                            expect(stateSlice.test2).to.equal(0);
                            break;
                        case 1:
                            expect(stateSlice.test2).to.equal(6);
                            break;
                        case 2:
                            expect(stateSlice.test2).to.equal(7);
                            break;
                    }
                    counter++;
                },
                e => {
                    expect(e).to.equal(null);
                },
                () => {
                }
            );

            setTimeout(()=> {
                store.dispatch({type: 'ACTION', body: 0});
            }, 300);

            setTimeout(()=> {
                store.dispatch({type: 'ACTION', body: 6});
            }, 400);

            setTimeout(()=> {
                store.dispatch({type: 'ACTION', body: 7});
            }, 500);

            setTimeout(()=> {
                expect(counter).to.equal(3);
                cb();
            }, 600);
        });
    });
});